package EM_LLR_Utils;

/*
 * Created on 10/10/2005
 *
 */

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import edu.tau.compbio.ds.SimilarityMatrix;
import edu.tau.compbio.graph.algo.MinimalSpanningTree;
import edu.tau.compbio.interaction.AnnotatedInteractorSet;
import edu.tau.compbio.interaction.Interaction;
import edu.tau.compbio.interaction.InteractionMap;
import edu.tau.compbio.interaction.Interactor;
import edu.tau.compbio.math.VecCalc;
import edu.tau.compbio.util.ProgressManager;

/**
 * @author ulitskyi
 *
 */
public class SeedGenerator<T extends AnnotatedInteractorSet> {
	public enum SeedConfinement {CONFINED_SUBSET,NON_CONFINED_SUBSET};
	public enum SeedScore {AVERAGE_SEED_SCORE,SUM_SEED_SCORE};
	public enum BestNeiSelection { ENUMERATION,HEURISTIC };
//	public enum SearchMethod {SIMPLE_GREEDY,OBSERVER_GREEDY};
	
	private InteractionMap _im;
	private SimilarityMatrix _wMat = null;
	private int[] _index2index = null;
	private ProgressManager _progManager = new ProgressManager();
	private SeedScore _scoreMethod = SeedScore.AVERAGE_SEED_SCORE;
	private SeedConfinement _seedMethod = SeedConfinement.NON_CONFINED_SUBSET;
	private BestNeiSelection _bestNeiMethod = BestNeiSelection.HEURISTIC;
	
	private boolean _printInfo = true;

	public SeedGenerator(InteractionMap im,SimilarityMatrix wMat,int[] index2index,SeedScore scoreMethod,SeedConfinement seedMethod)
	{
		_im = im;
		_wMat = wMat;
		_index2index = index2index;
		_scoreMethod = scoreMethod;
		_seedMethod = seedMethod;
	}
	
	public void setBestNeiSelectionMethod(BestNeiSelection method)
	{
		_bestNeiMethod = method;
	}
	
	protected AbstractList<Interactor> orderNodes()
	{
		float[] avs = new float[_wMat.size()];
		Iterator i = _im.nodesIterator();
		while (i.hasNext())
		{
			Interactor n = (Interactor) i.next();
			int ind = _index2index[n.getIndex()];
			if (ind==-1)
				continue;
			float score = 0;
			for (Interactor nei : (Set<Interactor>) _im.getAdjacentNodes(n))
			{
				int indNei = _index2index[nei.getIndex()];
				if (indNei==-1)
					continue;
				float w = _wMat.getSimilarity(ind,indNei);
				if (w>0)
					score+=w;
			}
			avs[ind] = score;
		}
		int[] ranks = VecCalc.sortWithRanks(avs);
		AbstractList<Interactor> outList = new ArrayList<Interactor>();
		for (int k=ranks.length-1;k>=0;k--)
		{
			String id = (String) _wMat.getObject(ranks[k]);
			Interactor n = _im.getInteractor(id); 
			if (n==null)
			{
				continue;
			}
			outList.add(n);
		}
		System.out.println("Finished ordering nodes");
		return outList;
	}
	public AbstractList<AnnotatedInteractorSet> generateAllPairsSeeds(int size,float minSimilarity,float minSeedScore,int maxSize)
	{
		System.out.println("Generating seeds...");
		AbstractList<AnnotatedInteractorSet> outList = new ArrayList<AnnotatedInteractorSet>();
		_progManager.resetProgress();
		// Try to find a seed from every node
		_im.clearColors();
		
		int count = 0;
		AbstractList<Interactor> order = orderNodes();
		for (Interactor node1 : order)
		{
			_progManager.incProgress(1f / _im.sizeInteractors());
			if (_printInfo && count++ % 50 == 0)
				System.out.println(count + "/"+order.size());
			if (node1.getColor()>0 || _index2index[node1.getIndex()]==-1)
				continue;
			ArrayList<Interactor> neis = getGoodNeis(node1,size-1,minSimilarity);
			if (neis==null)
				continue;
			neis.add(node1);
			HashSet<Interactor> seed = new HashSet<Interactor>();
			float bestScore = Float.NEGATIVE_INFINITY;
			if (_seedMethod == SeedConfinement.CONFINED_SUBSET && neis.size()>size)
			{
				bestScore = selectBestNeis(size,node1,neis,seed,minSimilarity);
			} else {
				AbstractList<Float> scores = null;
				AbstractList<Interactor> list = null;
				if (neis.size()>maxSize)
				{
					scores = new ArrayList<Float>();
					list = new ArrayList<Interactor>();
				}
				for (Interactor nei : neis)
				{
					int ind1 = _index2index[node1.getIndex()];
					int ind2 = _index2index[nei.getIndex()];
					float score = 0;
					if (ind1!=-1 && ind2!=-1)
					{
						score = _wMat.getSimilarity(ind1,ind2); 
						if (score<0 && nei!=node1)
							continue;
					}
					seed.add(nei);
					if (neis.size()>maxSize)
					{
						if (nei == node1)
							score = Float.MAX_VALUE;
						scores.add(score);
						list.add(nei);
					}						
				}
				if (seed.size()>maxSize)
				{
					seed.clear();
					float[] scoreArr = new float[scores.size()];
					for (int k=0;k<scoreArr.length;k++)
						scoreArr[k] = ((Float) scores.get(k)).floatValue();
					int[] ranks = VecCalc.sortWithRanks(scoreArr);
					for (int k=ranks.length-1;k>=ranks.length-maxSize;k--)
						seed.add(list.get(ranks[k]));
				}
				if (!seed.contains(node1))
				{
					throw new IllegalStateException("The seed does not contain its core!");
				}
				bestScore = testNodes(seed);
			}
			if (bestScore>minSeedScore)
			{
				Iterator j = seed.iterator();
				while (j.hasNext())
				{
					((Interactor) j.next()).setColor(1);
				}
				if (seed.size()>maxSize)
				{
					throw new IllegalStateException("Illegal seed size");
				}
				AnnotatedInteractorSet is = new AnnotatedInteractorSet(_im,String.valueOf(outList.size()+1),seed);
				is.setAttribute("Seed score",Float.valueOf(bestScore));
				is.setNodeProperty(Arrays.asList(new Interactor[] {node1}),"Seed core");
				outList.add(is);
			}
		}
		System.out.println("Finished generating seeds");
		return outList;
	}
	
	public AbstractList<T> generateAllPairsSeedsWithConfinedCut(int size,float minSimilarity,float minSeedScore,int maxSize,double maxCutWeight,Map<Interaction,Double> weights)
	{
		System.out.println("Generating seeds...");
		AbstractList<T> outList = new ArrayList<T>();
		_progManager.resetProgress();
		// Try to find a seed from every node
		_im.clearColors();
		
		
		MinimalSpanningTree mst = new MinimalSpanningTree(_im,weights);
		int count = 0;
		AbstractList<Interactor> order = orderNodes();
		for (Interactor starCenter : order)
		{
			_progManager.incProgress(1f / _im.sizeInteractors());
			if (_printInfo && count++ % 50 == 0)
				System.out.println(count + "/"+order.size());
			if (starCenter.getColor()>0 || _index2index[starCenter.getIndex()]==-1)
				continue;
			ArrayList<Interactor> neis = getGoodNeis(starCenter,size-1,minSimilarity);
			if (neis==null)
				continue;
			double curTotalWeight = 0;
			neis.add(starCenter);
			HashSet<Interactor> seed = new HashSet<Interactor>();
			seed.add(starCenter);
			float bestScore = Float.NEGATIVE_INFINITY;
			AbstractList<Float> scores = null;
			AbstractList<Interactor> list = null;
			if (neis.size()>maxSize)
			{
				scores = new ArrayList<Float>();
				list = new ArrayList<Interactor>();
			}
			boolean mstIsDirty = false;
			double prevMSTWeight = 0;
			for (Interactor nei : neis)
			{
				double w = 0;
				if (nei!=starCenter)
				{
					Interaction e = _im.getInteraction(starCenter,nei);
					w = weights.get(e);
					if (mstIsDirty && prevMSTWeight + w < maxCutWeight && curTotalWeight + w > maxCutWeight)
					{
						mst.findTree(seed);
						curTotalWeight=mst.getLastWeight();
						prevMSTWeight = curTotalWeight;
						mstIsDirty = false;
					} 
					if (curTotalWeight + w > maxCutWeight)
						continue;
				}
				int ind1 = _index2index[starCenter.getIndex()];
				int ind2 = _index2index[nei.getIndex()];
				float score = 0;
				if (ind1!=-1 && ind2!=-1)
				{
					score = _wMat.getSimilarity(ind1,ind2); 
					if (score<0 && nei!=starCenter)
						continue;
				}
				curTotalWeight+=w;
				seed.add(nei);
				mstIsDirty = true;
				if (neis.size()>maxSize)
				{
					if (nei == starCenter)
						score = Float.MAX_VALUE;
					scores.add(score);
					list.add(nei);
				}						
			}
			if (seed.size()>maxSize)
			{
				seed.clear();
				float[] scoreArr = new float[scores.size()];
				for (int k=0;k<scoreArr.length;k++)
					scoreArr[k] = ((Float) scores.get(k)).floatValue();
				int[] ranks = VecCalc.sortWithRanks(scoreArr);
				for (int k=ranks.length-1;k>=ranks.length-maxSize;k--)
					seed.add(list.get(ranks[k]));
			}
			if (!seed.contains(starCenter))
			{
				throw new IllegalStateException("The seed does not contain its core!");
			}
			bestScore = testNodes(seed);
			if (bestScore>minSeedScore)
			{
				Iterator j = seed.iterator();
				while (j.hasNext())
				{
					((Interactor) j.next()).setColor(1);
				}
				if (seed.size()>maxSize)
				{
					throw new IllegalStateException("Illegal seed size");
				}
				AnnotatedInteractorSet is = new AnnotatedInteractorSet(_im,String.valueOf(outList.size()+1),seed);
				is.setAttribute("Seed score",Float.valueOf(bestScore));
				is.setNodeProperty(Arrays.asList(new Interactor[] {starCenter}),"Seed core");
				outList.add((T)is);
			}
		}
		System.out.println("Finished generating seeds");
		return outList;
	}
	
	private float selectBestNeis(int size,Interactor node1,AbstractList<Interactor> neis,Collection<Interactor> seed,float minSimilarity)
	{
		float bestScore = Float.NEGATIVE_INFINITY;
		int[] inds = new int[size];
		int[] bestInds = new int[size];
		for (int k=0;k<size;k++)
		{
			inds[k] = k;
		}
		if (_bestNeiMethod == BestNeiSelection.ENUMERATION)
		{
			while (true) // do the enumeration
			{
				float score = testNodes(neis,inds,minSimilarity);
				if (!Float.isNaN(score) && (score > bestScore))
				{
					bestScore = score;
					for (int k=0;k<inds.length;k++)
						bestInds[k] = inds[k];
				}
				if (!updateIndices(inds,neis.size()))
					break;
			}
		} else if (_bestNeiMethod == BestNeiSelection.HEURISTIC) {
			float[] scores = new float[neis.size()];
			int count = 0;
			for (Interactor nei : neis)
			{
				scores[count++] = testNode(nei,neis);
			}
			seed.add(node1);
			int[] ranks = VecCalc.sortWithRanks(scores);
			for (int k=ranks.length-1;k>=0;k--)
			{
				if (seed.size()>size)
					break;
				seed.add(neis.get(ranks[k]));
			}
		}
		for (int k=0;k<bestInds.length;k++)
		{
			Interactor n = (Interactor) neis.get(bestInds[k]);
			seed.add(n);
		}
		return bestScore;
	}
	private ArrayList<Interactor> getGoodNeis(Interactor node,int minNumber,float minSimilarity)
	{
		Set<Interactor> neis = _im.getAdjacentNodes(node);
		if (neis.size()<minNumber)
			return null;
		Iterator<Interactor> i = neis.iterator();
		while (i.hasNext())
		{
			Interactor nei = i.next();
			if (nei.getColor()>0)
			{
				i.remove();
			}
		}
		int ind1 = _index2index[node.getIndex()];
		ArrayList<Interactor> goodNeis = new ArrayList<Interactor>();
		if (_wMat!=null && ind1!=-1)
		{
			for (Interactor nei : neis)
			{
				if (_seedMethod == SeedConfinement.CONFINED_SUBSET)
				{
					int ind2 = _index2index[nei.getIndex()];
					if (ind2!=-1)
					{
						float w = _wMat.getSimilarity(ind1,ind2);
						if (w<minSimilarity)
						{
							continue;
						}
					} else {
						continue;
					}
				}
				goodNeis.add(nei);
			}
		}
		return goodNeis;
	}
	private float testNodes(Set<Interactor> nodes)
	{
		ArrayList<Interactor> nodeList = new ArrayList<Interactor>(nodes);
		float sum = 0f;
		for (int i=0;i<nodeList.size();i++)
		{
			Interactor n1 = (Interactor) nodeList.get(i);
			int ind1 = _index2index[n1.getIndex()];
			if (ind1==-1)
				continue;
			for (int j=i+1;j<nodeList.size();j++)
			{
				Interactor n2 = (Interactor) nodeList.get(j);
				int ind2 = _index2index[n2.getIndex()];
				if (ind2==-1)
					continue;
				float corr = _wMat.getSimilarity(ind1,ind2);
				sum += corr;
			}
		}
		if (_scoreMethod == SeedScore.SUM_SEED_SCORE)
			return sum;
		else if (_scoreMethod == SeedScore.AVERAGE_SEED_SCORE)
			return sum / (nodeList.size() * (nodeList.size()-1) / 2);
		return Float.NaN;
	}
	private float testNode(Interactor n1,AbstractList<Interactor> nodes)
	{
		float sum = 0f;
		int ind1 = _index2index[n1.getIndex()];
		if (ind1 == -1)
			return sum;
		for (Interactor n2 : nodes)
		{
			if (n1 == n2)
				continue;
			int ind2 = _index2index[n2.getIndex()];
			if (ind2==-1)
				continue;
			float corr = _wMat.getSimilarity(ind1,ind2);
			if (corr > 0)
				sum += corr;
		}
		if (_scoreMethod == SeedScore.SUM_SEED_SCORE)
			return sum;
		else if (_scoreMethod == SeedScore.AVERAGE_SEED_SCORE)
			return sum / (nodes.size()* (nodes.size()-1) / 2);
		return Float.NaN;
	}
	private float testNodes(AbstractList<Interactor> nodes,int[] inds,float minSimilarity)
	{
		// Check the pairwise similarities between the indexed nodes
		float sum = 0f;
		for (int i=0;i<inds.length;i++)
		{
			Interactor n1 = (Interactor) nodes.get(i);
			int ind1 = _index2index[n1.getIndex()];
			if (ind1==-1)
				continue;
			for (int j=i+1;j<inds.length;j++)
			{
				Interactor n2 = (Interactor) nodes.get(j);
				int ind2 = _index2index[n2.getIndex()];
				if (ind2==-1)
					continue;
				float corr = _wMat.getSimilarity(ind1,ind2);
				sum += corr;
			}
		}
		if (_scoreMethod == SeedScore.SUM_SEED_SCORE)
			return sum;
		else if (_scoreMethod == SeedScore.AVERAGE_SEED_SCORE)
			return sum / (inds.length * (inds.length-1) / 2);
		return Float.NaN;
	}
	private boolean updateIndices(int[] inds,int maxInd)
	{
		int cur = inds.length-1;
		for (cur = inds.length-1;cur>=0;cur--)
		{
			int k = (inds.length-1) - cur; 
			if (inds[cur]<maxInd-k-1)
			{
				inds[cur]++;
				int count = inds[cur]+1; 
				for (int j=cur+1;j<inds.length;j++) // update the next indices
				{
					inds[j] = count++;
				}
				return true;
			}
		}
		return false;
	}
	
}

