package cplexAnalysis;

import java.util.HashSet;
import java.util.LinkedList;

import ilog.concert.IloException;
import ilog.concert.IloIntExpr;
import ilog.concert.IloIntVar;
import ilog.concert.IloNumExpr;
import ilog.cplex.IloCplex;
import ilog.cplex.IloCplex.UnknownObjectException;

/**
 * 
 * @author davidama
 * 
 *         This class uses the parsed data to perform a miqp analysis using
 *         cplex.java. The model of this class is the simplest, i.e. using
 *         average semantic similarity of genes to measure signatures
 *         similarity. The objective is to maximize all pair-wise
 *         similarities.That is, we are dealing here with one disease.
 * 
 */
public class CompleteMMFinder {

	/* Members */
	// g1 - the non-DC graph
	double[][] g1;
	// g2 the DC graph
	float[][] g2;
	// g1 threshold
	double g1_thr;
	// g2 threshold
	double g2_thr;

	IloCplex mainModel;
	// a variable i gets xi==1 if the gene i is in A
	IloIntVar[] xvars;
	// a variable i gets yi==1 if the gene i is in B
	IloIntVar[] yvars;

	// min size threshold
	int minModuleSize = 5;

	// results containers
	private double[] x_result;
	private double[] y_result;

	public CompleteMMFinder(double[][] a1, float[][] a2, double t1,
			double t2) {
		g1 = a1;
		g2 = a2;
		g1_thr = t1;
		g2_thr = t2;
		
		buildModel();
	}
	
	

	public void buildModel() {
		try {
			mainModel = new IloCplex();

			// we have integer variable for every gene that
			// can be selected in the signatures
			xvars = mainModel.boolVarArray(g1.length);
			yvars = mainModel.boolVarArray(g1.length);

			// generate the objective of the model
			// 1. add the consistency of assignment constrains
			IloNumExpr[] consistencyConstraints = new IloNumExpr[g1.length];
			for (int i = 0; i < consistencyConstraints.length; i++) {
				consistencyConstraints[i] = mainModel.sum(xvars[i],
						yvars[i]);
				mainModel.addLe(consistencyConstraints[i], 1);
			}

			// 2. add g1 constraints
			addG1Constraints(xvars);
			addG1Constraints(yvars);

			// add g2 constraints
			addG2Constraints();

			// add min A, B size constraints
			IloIntExpr xSum = mainModel.sum(xvars);
			mainModel.addGe(xSum, minModuleSize);
			IloIntExpr ySum = mainModel.sum(yvars);
			mainModel.addGe(ySum, minModuleSize);

			// add maximization criteria
			mainModel.addMaximize(mainModel.sum(xSum, ySum));

		} catch (IloException e) {
			// TODO Auto-generated catch block
//			 e.printStackTrace();
		}

		// solve the model
		try {
			mainModel.solve();

		} catch (IloException e) {
			System.out.println("no solution!");
		}
	}

	public void addG1Constraints(IloIntVar[] data2Xvars) throws IloException {
		for (int i = 0; i < g1.length; i++) {
			for (int j = 0; j < i; j++) {
				IloNumExpr xi_plus_xj = mainModel.sum(data2Xvars[i],
						data2Xvars[j]);
				int edgeVal = 0;
				if (g1[i][j] >= g1_thr) {
					edgeVal = 1;
				}
				edgeVal+=1;
				IloNumExpr val_minus_sol = mainModel.diff(edgeVal, xi_plus_xj);
				mainModel.addGe(val_minus_sol, 0);
			}
		}
	}

	public void addG2Constraints() throws IloException {
		for (int i = 0; i < g1.length; i++) {
			for (int j = 0; j < g1.length; j++) {
				if (i==j){
					continue;
				}
				IloNumExpr xi_plus_yj = mainModel.sum(xvars[i],
						yvars[j]);
				int edgeVal = 0;
				if (g2[i][j] >= g2_thr) {
					edgeVal = 1;
				}
				edgeVal+=1;
				IloNumExpr val_minus_sol = mainModel.diff(edgeVal, xi_plus_yj);
				mainModel.addGe(val_minus_sol, 0);
			}
		}
	}

	public void getXSolution() {

		if (x_result != null) {
			return;
		}

		x_result = new double[g1.length];
		for (int i = 0; i < x_result.length; i++) {
			try {
				x_result[i] = mainModel.getValue(xvars[i]);
			} catch (UnknownObjectException e) {
				// TODO Auto-generated catch block
//				e.printStackTrace();
			} catch (IloException e) {
				// TODO Auto-generated catch block
//				e.printStackTrace();
			}
		}

	}
	
	public void getYSolution() {

		if (y_result != null) {
			return;
		}

		y_result = new double[g1.length];
		for (int i = 0; i < y_result.length; i++) {
			try {
				y_result[i] = mainModel.getValue(yvars[i]);
			} catch (UnknownObjectException e) {
				// TODO Auto-generated catch block
//				e.printStackTrace();
			} catch (IloException e) {
				// TODO Auto-generated catch block
//				e.printStackTrace();
			}
		}

	}
	
	public LinkedList<HashSet<Integer>> getSelectedMetaModule(){
		LinkedList<HashSet<Integer>> sol = new LinkedList<HashSet<Integer>>();
		
		getXSolution();
		HashSet<Integer> x = new HashSet<Integer>();
		for (int i = 0; i < x_result.length; i++) {
			if (x_result[i]>=1){
				x.add(i);
			}
		}
		
		getYSolution();
		HashSet<Integer> y = new HashSet<Integer>();
		for (int i = 0; i < y_result.length; i++) {
			if (y_result[i]>=1){
				y.add(i);
			}
		}
		
		sol.add(x);
		System.out.println("x size is: "+x.size());
		sol.add(y);
		
		return sol;
	}
}
