package graph_algorithms;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

public abstract class AbstractMetaModuleFinder {
	/* MEMBERS */

	/*
	 * We use CLICK model to create a log-likelihood matrix. We repeat the CLICK
	 * process twice. The first is on the global similarity matrix and requires
	 * an EM step (not done here) to find the mates and non mates distributions.
	 * The second matrix is created by analyzing the results of running our
	 * differential correlation (DC) analysis on the real data and comparing it
	 * to DC analysis on shuffled data sets.
	 */
	float[][] consistentCorrelationsGraph;
	public float[][] DC_graph;

	float likelihoodThresholdForDC = 0;
	float likelihoodThresholdForCG = 0;
	double sizesConsistencyThr = 0.25;

	// this array holds the singletons
	// if a gene is a singleton than it was not used
	protected boolean[] wasUsed;

	/* Other Auxiliary MEMBERS */
	public ArrayList<String> geneNames;
	public static int minSeedModuleSize = 2;
	
	public static int minModuleSize=2;

	/* These objects shall contain the final results */
	LinkedList<HashSet<Integer>> modules;
	short[][] areModulesFriends;
	LinkedList<MetaModule> meta_modules;
	
	public void printModulesAndBiModulesStatistics() {
		// get some statistics 
		int numCoveredGenes = 0;
		int maxMMSize = 0;
		for (Iterator<MetaModule> iterator = meta_modules.iterator(); iterator.hasNext();) {
			MetaModule mm = (MetaModule) iterator.next();
			int currSize =0;
			currSize += mm.U.size();
			currSize += mm.V.size();
			numCoveredGenes += currSize;
			if (currSize>maxMMSize){
				maxMMSize = currSize;
			}
		}
		System.out.println("num modules: " + modules.size());
		System.out.println("num bi-modules: " + meta_modules.size());
		System.out.println("num covered genes: " + numCoveredGenes);
		System.out.println("maximal mm size: " + maxMMSize);
	}
	
	public void printModules(String prefix, String path, int minSize) throws IOException {

		BufferedWriter writer = new BufferedWriter(new FileWriter(path));
		int counter = 0;
		for (Iterator<HashSet<Integer>> iterator = modules.iterator(); iterator
				.hasNext();) {
			HashSet<Integer> module = (HashSet<Integer>) iterator.next();
			if (module.size() < minSize) {
				continue;
			}
			for (Iterator<Integer> iterator2 = module.iterator(); iterator2
					.hasNext();) {
				Integer geneIndex = (Integer) iterator2.next();
				String geneName = geneNames.get(geneIndex);
				String seedName = prefix + "_" + counter;
				writer.write(geneName + "\t" + seedName + "\n");
			}
			counter++;
		}
		writer.close();
	}

	public void printModules(String prefix, BufferedWriter writer, int minSize,
			Hashtable<String, HashSet<String>> seeds) {

		try {
			int counter = 0;
			for (Iterator<HashSet<Integer>> iterator = modules.iterator(); iterator
					.hasNext();) {
				HashSet<Integer> module = (HashSet<Integer>) iterator.next();
				if (module.size() < minSize) {
					continue;
				}
				for (Iterator<Integer> iterator2 = module.iterator(); iterator2
						.hasNext();) {
					Integer geneIndex = (Integer) iterator2.next();
					String geneName = geneNames.get(geneIndex);
					String seedName = prefix + "_" + counter;
					if (minSize <= 1) {
						if (!seeds.containsKey(seedName)) {
							HashSet<String> set = new HashSet<String>();
							set.add(geneName);
							// seeds.put(seedName, set);
						} else {
							seeds.get(seedName).add(geneName);
						}
					}

					writer.write(geneName + "\t" + seedName + "\n");
				}
				
				counter++;
			}
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	public void printMetaGraph(String prefix, String path) throws IOException{
		BufferedWriter writer = new BufferedWriter(new FileWriter(path));
		
		for (int i = 0; i < areModulesFriends.length; i++) {
			for (int j = 0; j < i; j++) {
				if (areModulesFriends[i][j]==1){
					String seed1Name = prefix + "_" + i;
					String seed2Name = prefix + "_" + j;
					writer.write(seed1Name + "\tEdge\t" + seed2Name + "\n");
				}
			}
		}
		writer.close();
	}

	public void printMetaClusters(String prefix, String path,
			Hashtable<String, HashSet<String>> mms) throws IOException {
		int countModules = -1;

		BufferedWriter writer = new BufferedWriter(new FileWriter(path));

		for (Iterator<MetaModule> iterator = meta_modules.iterator(); iterator
				.hasNext();) {
			MetaModule metaCluster = (MetaModule) iterator.next();
			countModules++;

			for (Iterator<Integer> iterator2 = metaCluster.getUnion()
					.iterator(); iterator2.hasNext();) {
				Integer index = (Integer) iterator2.next();
				String gene = geneNames.get(index);
				String clusterName = prefix + "_" + countModules + "_" + "_"
						+ metaCluster.betweenLLRScore;
				if (!mms.containsKey(clusterName)) {
					HashSet<String> mm = new HashSet<String>();
					mm.add(gene);
					// mms.put(clusterName, mm);
				} else {
					mms.get(clusterName).add(gene);
				}
				writer.write(gene + "\t" + clusterName + "\n");
			}
		}
		writer.close();

	}

	public void printMetaClusters(String prefix, BufferedWriter writer,
			Hashtable<String, HashSet<String>> mms) {
		int countModules = -1;
		try {
			for (Iterator<MetaModule> iterator = meta_modules.iterator(); iterator
					.hasNext();) {
				MetaModule metaCluster = (MetaModule) iterator.next();
				countModules++;

				for (Iterator<Integer> iterator2 = metaCluster.getUnion()
						.iterator(); iterator2.hasNext();) {
					Integer index = (Integer) iterator2.next();
					String gene = geneNames.get(index);
					String clusterName = prefix + "_" + countModules + "_"
							+ "_" + metaCluster.betweenLLRScore;
					if (!mms.containsKey(clusterName)) {
						HashSet<String> mm = new HashSet<String>();
						mm.add(gene);
						// mms.put(clusterName, mm);
					} else {
						mms.get(clusterName).add(gene);
					}

					writer.write(gene + "\t" + clusterName + "\n");
				}
			}
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}

	public void setMinSeedModuleSize(int minSeedModuleSize) {
		AbstractMetaModuleFinder.minSeedModuleSize = minSeedModuleSize;
	}

	public void setSizesConsistencyThr(double sizesConsistencyThr) {
		this.sizesConsistencyThr = sizesConsistencyThr;
	}
	
	
	/**
	 * Static auxiliary methods
	 * 
	 */

	public static void sortEdgesByScores(LinkedList<ComparableEdge> edges,
			Hashtable<Integer, HashSet<Integer>> neighborsListsInDCAGraph) {
		LinkedList<ComparableEdge> toRemove = new LinkedList<ComparableEdge>();
		for (Iterator<ComparableEdge> iterator = edges.iterator(); iterator
				.hasNext();) {
			ComparableEdge e = (ComparableEdge) iterator.next();
			int u = e.u;
			int v = e.v;
			double deg1 = neighborsListsInDCAGraph.get(u).size();
			double deg2 = neighborsListsInDCAGraph.get(v).size();
			e.score = deg1 + deg2;
			if (e.score == 0) {
				toRemove.add(e);
			} else {
				e.score *= Math.min(deg1 / deg2, deg2 / deg1);
			}
		}

		edges.removeAll(toRemove);
		Collections.sort(edges);
	}
	
	public static void removeCurrentListFromGraph(
			Hashtable<Integer, HashSet<Integer>> neighborsListsInDCAGraph,
			HashSet<Integer> set) {

		for (Iterator<Integer> iterator2 = set.iterator(); iterator2.hasNext();) {
			Integer inSet = (Integer) iterator2.next();
			neighborsListsInDCAGraph.put(inSet, new HashSet<Integer>());
		}

		for (Iterator<Integer> iterator = neighborsListsInDCAGraph.keySet()
				.iterator(); iterator.hasNext();) {
			Integer index = (Integer) iterator.next();
			HashSet<Integer> list = neighborsListsInDCAGraph.get(index);
			if (list.size() == 0) {
				continue;
			}
			list.removeAll(set);
		}
	}

	public static void removeIntersection(HashSet<Integer> neighborsOfU,
			HashSet<Integer> neighborsOfV) {

		LinkedList<Integer> intersection = new LinkedList<Integer>();
		for (Iterator<Integer> iterator = neighborsOfU.iterator(); iterator
				.hasNext();) {
			Integer neighborOfU = (Integer) iterator.next();
			if (neighborsOfV.contains(neighborOfU)) {
				intersection.add(neighborOfU);
			}
		}
		neighborsOfU.removeAll(intersection);
		neighborsOfV.removeAll(intersection);
	}
	
	public void removeZeroDegreesMetaNodes() {
		LinkedList<HashSet<Integer>> nModules;
		int[] degrees = new int[modules.size()];
		for (int i = 0; i < areModulesFriends.length; i++) {
			for (int j = 0; j < i; j++) {
				if (areModulesFriends[i][j]==1) {
					degrees[i]++;
					degrees[j]++;
				}
			}
		}

		nModules = new LinkedList<HashSet<Integer>>();
		for (int i = 0; i < degrees.length; i++) {
			if (degrees[i] > 0) {
				nModules.add(modules.get(i));
			}
		}
		modules = nModules;
		calculateMetaGraph();
	}

	public void setLikelihoodThresholdForDC(float likelihoodThresholdForDCA) {
		this.likelihoodThresholdForDC = likelihoodThresholdForDCA;
	}
	
	public void calculateMetaGraph() {

		areModulesFriends = new short[modules.size()][modules.size()];
		for (int i = 0; i < areModulesFriends.length; i++) {
			HashSet<Integer> firstModule = modules.get(i);
			for (int j = 0; j < i; j++) {
				HashSet<Integer> secondModule = modules.get(j);
				double score = 0;
				for (Iterator<Integer> iterator = secondModule.iterator(); iterator
						.hasNext();) {
					Integer node1 = (Integer) iterator.next();

					for (Iterator<Integer> iterator2 = firstModule.iterator(); iterator2
							.hasNext();) {
						Integer node2 = (Integer) iterator2.next();
						score += DC_graph[node1][node2];
					}

				}

				if (score > likelihoodThresholdForDC) {
					areModulesFriends[i][j] = 1;
					areModulesFriends[j][i] = 1;
				}
			}
		}

	}
	
	public void markAsUsed(HashSet<Integer> neighborsOfU) {
		for (Iterator<Integer> iterator = neighborsOfU.iterator(); iterator
				.hasNext();) {
			Integer neighbor = (Integer) iterator.next();
			wasUsed[neighbor] = true;
		}
	}
	
	public List<HashSet<Integer>> getModulesCopy() {
		List<HashSet<Integer>> copy = new ArrayList<HashSet<Integer>>(modules.size());
		for (Iterator<HashSet<Integer>> iterator = modules.iterator(); iterator.hasNext();) {
			HashSet<Integer> m = (HashSet<Integer>) iterator.next();
			copy.add(new HashSet<Integer>(m));
		}
		return copy;
	}
	
	public void setModules(List<HashSet<Integer>> modules) {
		this.modules = new LinkedList<HashSet<Integer>>();
		for (Iterator<HashSet<Integer>> iterator = modules.iterator(); iterator.hasNext();) {
			HashSet<Integer> m = (HashSet<Integer>) iterator.next();
			this.modules.add(new HashSet<Integer>(m));
		}
	}
	
	public short[][] getAreModulesFriends() {
		return areModulesFriends;
	}
	
	public LinkedList<MetaModule> getMeta_modules() {
		return meta_modules;
	}
}
