package graph_algorithms;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;

public class MetaModuleFinder extends AbstractMetaModuleFinder{

	/**
	 * Constructor
	 * 
	 * @param c
	 *            - the consistent correlations graph
	 * @param dc
	 *            - the DC analysis results
	 * @param ids
	 *            - the gene names
	 */
	public MetaModuleFinder(float[][] c, float[][] dc, ArrayList<String> ids) {
		DC_graph = dc;
		consistentCorrelationsGraph = c;
		geneNames = ids;
	}

	/**
	 * This methods receives as input two distinct sets. The set elements
	 * represent the index of a gene in the matrices. The algorithm greedily
	 * tries to remove a gene from T or S until the remaining sets satisfy the
	 * meta-module definition.
	 * 
	 * @param T
	 * @param S
	 * @return
	 */
	public boolean greedyRemoveNodesFromTwoSides(Collection<Integer> T,
			Collection<Integer> S) {

		HashSet<Integer> toRemoveS = new HashSet<Integer>();
		for (Iterator<Integer> iterator = S.iterator(); iterator.hasNext();) {
			Integer inS = (Integer) iterator.next();
			int currDegree = 0;
			for (Iterator<Integer> iterator2 = T.iterator(); iterator2
					.hasNext();) {
				Integer inT = (Integer) iterator2.next();
				if (DC_graph[inS][inT] > 0) {
					currDegree++;
				}
			}
			if (currDegree <= minSeedModuleSize-1) {
				toRemoveS.add(inS);
			}
		}

		HashSet<Integer> toRemoveT = new HashSet<Integer>();
		for (Iterator<Integer> iterator = T.iterator(); iterator.hasNext();) {
			Integer inT = (Integer) iterator.next();
			int currDegree = 0;
			for (Iterator<Integer> iterator2 = S.iterator(); iterator2
					.hasNext();) {
				Integer inS = (Integer) iterator2.next();
				if (DC_graph[inS][inT] > 0) {
					currDegree++;
				}
			}
			if (currDegree <= minSeedModuleSize-1) {
				toRemoveT.add(inT);
			}
		}

		S.removeAll(toRemoveS);
		T.removeAll(toRemoveT);

		// System.out.println("cleaning stage removed: " +
		// toRemoveS.size()+toRemoveT.size());

		Hashtable<Integer, Double> inTNode2WithinScore = getWithinSetScores(T);
		Hashtable<Integer, Double> inSNode2WithinScore = getWithinSetScores(S);

		Hashtable<Integer, Double> fromTtoSScores = new Hashtable<Integer, Double>();
		Hashtable<Integer, Double> fromStoTScores = new Hashtable<Integer, Double>();

		for (Iterator<Integer> iterator = S.iterator(); iterator.hasNext();) {
			Integer inS = (Integer) iterator.next();
			fromStoTScores.put(inS, 0.0);
			for (Iterator<Integer> iterator2 = T.iterator(); iterator2
					.hasNext();) {
				Integer inT = (Integer) iterator2.next();
				if (!fromTtoSScores.containsKey(inT)) {
					fromTtoSScores.put(inT, 0.0);
				}
				fromStoTScores.put(inS, fromStoTScores.get(inS)
						+ DC_graph[inS][inT]);

				fromTtoSScores.put(inT, fromTtoSScores.get(inT)
						+ DC_graph[inS][inT]);
			}
		}

		while (true) {
			
			if (T.size() <= minSeedModuleSize || S.size() <= minSeedModuleSize) {
				return false;
			}

			double minScore = Double.MAX_VALUE;
			int minCombScoreIndex = -1;
			double min1 = minScore;
			double min2 = minScore;
			double minBetween = Double.MAX_VALUE;
			int minBetweenIndex = -1;
			double minWithin = Double.MAX_VALUE;
			int minWithinIndex = -1;
			int minNodeSelected = -1;

			for (Iterator<Integer> iterator = T.iterator(); iterator.hasNext();) {
				Integer node = (Integer) iterator.next();
				if (inTNode2WithinScore.get(node) + fromTtoSScores.get(node) < minScore) {
					minScore = inTNode2WithinScore.get(node)
							+ fromTtoSScores.get(node);
					min1 = inTNode2WithinScore.get(node);
					min2 = fromTtoSScores.get(node);
					minCombScoreIndex = node;
				}
				if (inTNode2WithinScore.get(node) < minWithin) {
					minWithin = inTNode2WithinScore.get(node);
					minWithinIndex = node;
				}
				if (fromTtoSScores.get(node) < minBetween) {
					minBetween = fromTtoSScores.get(node);
					minBetweenIndex = node;
				}
			}

			for (Iterator<Integer> iterator = S.iterator(); iterator.hasNext();) {
				Integer node = (Integer) iterator.next();
				if (inSNode2WithinScore.get(node) + fromStoTScores.get(node) < minScore) {
					minScore = inSNode2WithinScore.get(node)
							+ fromStoTScores.get(node);
					min1 = inSNode2WithinScore.get(node);
					min2 = fromStoTScores.get(node);
					minNodeSelected = node;
					minCombScoreIndex = node;
				}
				if (inSNode2WithinScore.get(node) < minWithin) {
					minWithin = inSNode2WithinScore.get(node);
					minWithinIndex = node;
				}
				if (fromStoTScores.get(node) < minBetween) {
					minBetween = fromStoTScores.get(node);
					minBetweenIndex = node;
				}
			}
			// if (T.size()==1 && S.size()==1 && minBetween > 0){
			// return true;
			// }

			if (min1 <= likelihoodThresholdForDC
					&& min2 <= likelihoodThresholdForCG) {
				minNodeSelected = minCombScoreIndex;
			} else if (min1 <= likelihoodThresholdForDC) {
				minNodeSelected = minWithinIndex;
			} else {
				minNodeSelected = minBetweenIndex;
			}

			if (min1 > likelihoodThresholdForCG
					&& min2 > likelihoodThresholdForDC
					&& minBetween > likelihoodThresholdForDC
					&& minWithin > likelihoodThresholdForCG) {
				return true;
			}

			if (T.contains(minNodeSelected)) {

				inTNode2WithinScore.remove(minNodeSelected);
				T.remove(minNodeSelected);
				fromTtoSScores.remove(minNodeSelected);

				for (Iterator<Integer> iterator2 = T.iterator(); iterator2
						.hasNext();) {
					Integer inT = (Integer) iterator2.next();
					double score = inTNode2WithinScore.get(inT);
					score -= consistentCorrelationsGraph[inT][minNodeSelected];
					inTNode2WithinScore.put(inT, score);
				}

				for (Iterator<Integer> iterator2 = S.iterator(); iterator2
						.hasNext();) {
					Integer inS = (Integer) iterator2.next();
					double score = fromStoTScores.get(inS);
					score -= DC_graph[inS][minNodeSelected];
					fromStoTScores.put(inS, score);
				}

			} else if (S.contains(minNodeSelected)) {

				inSNode2WithinScore.remove(minNodeSelected);
				S.remove(minNodeSelected);
				fromStoTScores.remove(minNodeSelected);

				for (Iterator<Integer> iterator2 = S.iterator(); iterator2
						.hasNext();) {
					Integer inS = (Integer) iterator2.next();
					double score = inSNode2WithinScore.get(inS);
					score -= consistentCorrelationsGraph[inS][minNodeSelected];
					inSNode2WithinScore.put(inS, score);
				}

				for (Iterator<Integer> iterator2 = T.iterator(); iterator2
						.hasNext();) {
					Integer inT = (Integer) iterator2.next();
					double score = fromTtoSScores.get(inT);
					score -= DC_graph[inT][minNodeSelected];
					fromTtoSScores.put(inT, score);
				}
			}
		}
	}

	public Hashtable<Integer, Double> getWithinSetScores(Collection<Integer> T) {

		Hashtable<Integer, Double> node2Score = new Hashtable<Integer, Double>();
		for (Iterator<Integer> iterator = T.iterator(); iterator.hasNext();) {
			Integer inT = (Integer) iterator.next();
			node2Score.put(inT, 0.0);
			for (Iterator<Integer> iterator2 = T.iterator(); iterator2
					.hasNext();) {
				Integer secondInT = (Integer) iterator2.next();
				if (inT == secondInT) {
					continue;
				}
				double score = node2Score.get(inT);
				score += consistentCorrelationsGraph[inT][secondInT];
				node2Score.put(inT, score);
			}
		}
		return node2Score;
	}

	public void findSeedPairs(int minSeedSize) {

		modules = new LinkedList<HashSet<Integer>>();
		meta_modules = new LinkedList<MetaModule>();

		wasUsed = new boolean[DC_graph.length];
		Arrays.fill(wasUsed, false);

		// get all edges from the DC process
		LinkedList<ComparableEdge> edges = getEdgesForAlgo(likelihoodThresholdForDC);
//		double thr = likelihoodThresholdForDC;
//		while (edges.size() > 800000) {
//			thr++;
//			edges = getEdgesForAlgo(thr);
//		}

		Hashtable<Integer, HashSet<Integer>> neighborsListsInDCAGraph = createNeighborsListsInGraph();
		sortEdgesByScores(edges, neighborsListsInDCAGraph);

		int itersCounter = 0;
		while (edges.size() > 0 && itersCounter < 800000) {
			itersCounter++;

			ComparableEdge currentEdge = edges.removeLast();
			int u = currentEdge.u;
			int v = currentEdge.v;
			if (wasUsed[u] || wasUsed[v]) {
				continue;
			}

			// get initial neighbors lists
			HashSet<Integer> neighborsOfU = new HashSet<Integer>(
					neighborsListsInDCAGraph.get(u));
			HashSet<Integer> neighborsOfV = new HashSet<Integer>(
					neighborsListsInDCAGraph.get(v));

			if (neighborsOfU.size() < minSeedSize
					|| neighborsOfV.size() < minSeedSize) {
				continue;
			}

			removeIntersection(neighborsOfU, neighborsOfV);

			if (neighborsOfU.size() < minSeedSize
					|| neighborsOfV.size() < minSeedSize) {
				continue;
			}

			greedyRemoveNodesFromTwoSides(neighborsOfU, neighborsOfV);

			if (neighborsOfU.size() < minSeedSize
					|| neighborsOfV.size() < minSeedSize) {
				// System.out.println();
				continue;
			}

			double s1 = neighborsOfU.size();
			double s2 = neighborsOfV.size();

			if (s1 >= minSeedSize && s2 >= minSeedSize
					&& ((Math.min(s1 / s2, s2 / s1)) >= sizesConsistencyThr)) {

				modules.add(new HashSet<Integer>(neighborsOfV));
				modules.add(new HashSet<Integer>(neighborsOfU));
				MetaModule currPair = new MetaModule(new HashSet<Integer>(
						neighborsOfU), new HashSet<Integer>(neighborsOfV), this);
				meta_modules.add(currPair);

				// System.out.print(" found pair: " + neighborsOfU.size() + ","
				// + neighborsOfV.size() +" score is " +
				// currPair.betweenLLRScore +"\n");

				// go over all covered genes and mark as used
				markAsUsed(neighborsOfU);
				markAsUsed(neighborsOfV);

				// removing u and v from the graph
				removeCurrentListFromGraph(neighborsListsInDCAGraph,
						neighborsOfU);
				removeCurrentListFromGraph(neighborsListsInDCAGraph,
						neighborsOfV);
				// neighborsListsInDCAGraph = createNeighborsListsInGraph();

				// update edges lists
				sortEdgesByScores(edges, neighborsListsInDCAGraph);

			} else {
				// System.out.println();
			}
		}
	}


	public Hashtable<Integer, HashSet<Integer>> createNeighborsListsInGraph() {
		Hashtable<Integer, HashSet<Integer>> neighborsListsInDCAGraph = new Hashtable<Integer, HashSet<Integer>>();
		for (int i = 0; i < DC_graph.length; i++) {
			neighborsListsInDCAGraph.put(i, new HashSet<Integer>());
		}

		for (int i = 0; i < DC_graph.length; i++) {
			if (wasUsed[i]) {
				continue;
			}
			for (int j = 0; j < i; j++) {
				if (wasUsed[j]) {
					continue;
				}
				if (DC_graph[i][j] > likelihoodThresholdForDC) {
					neighborsListsInDCAGraph.get(i).add(j);
					neighborsListsInDCAGraph.get(j).add(i);
				}
			}
		}
		return neighborsListsInDCAGraph;
	}

	public LinkedList<ComparableEdge> getEdgesForAlgo(
			double likelihoodThresholdForDCA) {
		LinkedList<ComparableEdge> edges = new LinkedList<ComparableEdge>();
		for (int i = 0; i < DC_graph.length; i++) {
			for (int j = 0; j < i; j++) {
				if (DC_graph[i][j] > likelihoodThresholdForDCA) {
					ComparableEdge e = new ComparableEdge(i, j, 0);
					edges.add(e);
				}
			}
		}
		return edges;
	}

	public LinkedList<LinkedList<Integer>> addSingletonsToModules() {

		boolean[] wasUsed = new boolean[DC_graph.length];
		for (Iterator<HashSet<Integer>> iterator = modules.iterator(); iterator
				.hasNext();) {
			HashSet<Integer> m = (HashSet<Integer>) iterator.next();
			for (Iterator<Integer> iterator2 = m.iterator(); iterator2
					.hasNext();) {
				Integer inm = (Integer) iterator2.next();
				wasUsed[inm] = true;
			}
		}

		ArrayList<Integer> sigs = new ArrayList<Integer>();
		for (int i = 0; i < wasUsed.length; i++) {
			if (!wasUsed[i]) {
				sigs.add(i);
			}
		}

		SingletonsAndMetaModulesMerger merger = new SingletonsAndMetaModulesMerger(
				meta_modules, sigs, this);
		merger.doMerges(1, 0);
		meta_modules = new LinkedList<MetaModule>();
		for (int i = 0; i < merger.pairs.length; i++) {
			meta_modules.add(merger.pairs[i]);
		}

		modules = merger.getModulesList();
		return merger.getMetaModules();
	}

	public void getFinalSolution(int minSize) {
		findSeedPairs(minSeedModuleSize);
		improveSolution(minSize);
	}

	public void improveSolution(int minSize) {
		// merge seeds
		MetaModulesMerger merger = new MetaModulesMerger(meta_modules);
		merger.doMerges();
		meta_modules = new LinkedList<MetaModule>();
		for (int i = 0; i < merger.pairs.length; i++) {
			meta_modules.add(merger.pairs[i]);
		}
		modules = merger.getModulesList();

		// add singletons
		addSingletonsToModules();
		Collections.sort(meta_modules);
		
		modules = new LinkedList<HashSet<Integer>>();
		LinkedList<MetaModule> newPairs = new LinkedList<MetaModule>();
		for (int i = meta_modules.size() - 1; i >= 0; i--) {
			if (meta_modules.get(i).getU().size() > minSize) {
				modules.add(meta_modules.get(i).U);
			}
			if (meta_modules.get(i).getV().size() > minSize) {
				modules.add(meta_modules.get(i).V);
			}
			if (meta_modules.get(i).getUnion().size() > 2*minSize) {
				newPairs.add(meta_modules.get(i));
			}
		}

		meta_modules = newPairs;
		calculateMetaGraph();
	}

	public void getFinalSolution() {
		getFinalSolution(-Integer.MAX_VALUE);
	}
}
