package math_utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

/**
 * This class implements and expression matrix object.
 * It supports parsing of an expression matrix file and manipulations
 * of the data.
 *
 */

public class ExpressionMatrix {
	public float[][] matrix;
	String[] header;
	public ArrayList<String> genesID=new ArrayList<String>();
	private int genesCounter=0;
	Hashtable<String, Integer> geneName2Index = new Hashtable<String, Integer>();
	
	public ExpressionMatrix(float[][] mat,String[] h,ArrayList<String> ids) {
		matrix = mat;
		header = h;
		genesID = ids;
	}
	
	public ExpressionMatrix(String path) {
		try {
			BufferedReader filebuffer = new BufferedReader(new FileReader(path));
			String currLine;
			//header processing
			currLine = filebuffer.readLine();
			if (currLine.equals("")){
				System.err.println("please provide a valid matrix file");
				System.exit(1);
			}
			String[] headerTmp = currLine.split("\t");
			header = new String[headerTmp.length-1];
			for (int i = 1; i < headerTmp.length; i++) {
				header[i-1] = headerTmp[i];
			}
			
						
			ArrayList<ArrayList<Float>> tmpMatrix = new ArrayList<ArrayList<Float>>();
			int linesCounter =1;
			while((currLine=filebuffer.readLine()) != null){
				linesCounter++;
				String[] gene = currLine.split("\t");
				if (gene.length != header.length+1){
					System.out.println("continue : " + linesCounter);
					continue;
				}
				ArrayList<Float> geneVals = new ArrayList<Float>();
				genesID.add(gene[0]);
				for (int i = 1; i < gene.length; i++) {
					geneVals.add(Float.parseFloat(gene[i]));
				}
				tmpMatrix.add(geneVals);
				genesCounter++;
			}
			filebuffer.close();
			matrix = new float[tmpMatrix.size()][tmpMatrix.get(0).size()];
			for (int i = 0; i < tmpMatrix.size(); i++) {
				for (int j = 0; j < tmpMatrix.get(0).size(); j++) {
					matrix[i][j] = tmpMatrix.get(i).get(j);
				}
//				System.out.println(matrix[i][0]);
			}
			
			tmpMatrix = null;
			Runtime.getRuntime().gc();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}	
	}
	
	/**
	 * unite clusters into median pattern
	 * @param clusteringResult
	 * @return
	 */
	public ExpressionMatrix matrixByClusteringResults(LinkedList<LinkedList<Integer>> clusteringResult){

		int size = clusteringResult.size();
		float[][] newMatrix = new float[size][header.length];

		int clustersCounter = 0;
		ArrayList<String> newGeneIds = new ArrayList<String>();
		for (Iterator<LinkedList<Integer>> iterator = clusteringResult.iterator(); iterator.hasNext();) {
			LinkedList<Integer> cluster = (LinkedList<Integer>) iterator.next();
			int counter=0;
			for (Iterator<Integer> iterator2 = cluster.iterator(); iterator2.hasNext();) {
				Integer gene = (Integer) iterator2.next();
				for (int j = 0; j < matrix[0].length; j++) {
					newMatrix[clustersCounter][j]+=matrix[gene][j];
				}
				counter++;
				for (int j = 0; j < matrix[0].length; j++) {
					newMatrix[clustersCounter][j] /= cluster.size();
				}
			}
			if (cluster.size()>1){
				newGeneIds.add("cluster_"+clustersCounter + "_"+System.currentTimeMillis());
				
			} 
			else{
				newGeneIds.add(genesID.get(cluster.get(0) ) + "_"+System.currentTimeMillis());
			}
			clustersCounter++;
			
		}
		return new ExpressionMatrix(newMatrix,header,newGeneIds);
	}
	
	public ExpressionMatrix matrixByClusteringResults(List<HashSet<Integer>> clusteringResult){

		int size = clusteringResult.size();
		float[][] newMatrix = new float[size][header.length];

		int clustersCounter = 0;
		ArrayList<String> newGeneIds = new ArrayList<String>();
		for (Iterator<HashSet<Integer>> iterator = clusteringResult.iterator(); iterator.hasNext();) {
			HashSet<Integer> cluster = (HashSet<Integer>) iterator.next();
			int counter=0;
			for (Iterator<Integer> iterator2 = cluster.iterator(); iterator2.hasNext();) {
				Integer gene = (Integer) iterator2.next();
				for (int j = 0; j < matrix[0].length; j++) {
					newMatrix[clustersCounter][j]+=matrix[gene][j];
				}
				counter++;
				for (int j = 0; j < matrix[0].length; j++) {
					newMatrix[clustersCounter][j] /= cluster.size();
				}
			}
			newGeneIds.add("cluster_"+clustersCounter + "_"+System.currentTimeMillis());
			clustersCounter++;
			
		}
		return new ExpressionMatrix(newMatrix,header,newGeneIds);
	}
	
	/**
	 * This method returns a sub-matrix with all the
	 * raws that appear in the list.
	 * The given parameter is an ArrayList of indices
	 * of raws in the matrix, in ascending order;
	 */
	public ExpressionMatrix getSubMatrixByRowsIndices(LinkedList<Integer> list){
		float[][] newMatrix = new float [list.size()][matrix[0].length];
		ArrayList<String> newGenesID = new ArrayList<String>();
		int i=0;
		for (Iterator<Integer> iterator = list.iterator(); iterator.hasNext();) {
			Integer gene = (Integer) iterator.next();
			for (int j = 0; j < newMatrix[0].length; j++) {
				newMatrix[i][j]=matrix[gene][j];
			}
			newGenesID.add(genesID.get(gene));
			i++;
		}
		
		return new ExpressionMatrix(newMatrix,header,newGenesID);
	}
	
	/**
	 * This method receives a sorted list of indices
	 * and returns a sub-matrix that contains only the
	 * columns that are included in the list
	 */
	public ExpressionMatrix getSubMatrixByColumnsIndices(ArrayList<Integer> list){
		int size = list.size();
		float[][] newMatrix = new float[matrix.length][size];
		String[] newHeader = new String[size];
		for (int j = 0; j < size; j++) {
			int currIndex = list.get(j);
			for (int i = 0; i < matrix.length; i++) {
				newMatrix[i][j]=matrix[i][currIndex];  
			}
			newHeader[j]=header[currIndex];		
		}
		
		
		return new ExpressionMatrix(newMatrix, newHeader, genesID);
	}
	
	/**
	 * WEKA receives tab separated files, the header contains the features names
	 * and each line is an instance.
	 * Therefore the expander matrix is the transpose of the WEKA's
	 * matrix. This method prints the transposed matrix. 
	 * The last column in the output matrix is the class name, this is
	 * the default in WEKA.
	 * @param writer
	 */
	public void printWekaCSVInstancesArePatients(String path,int[] borders,int classIndex){
		//print header
		
		BufferedWriter writer;
		try {
			writer = new BufferedWriter(new FileWriter(path));
			writer.write(genesID.get(0));
			for (int i = 1; i < matrix.length; i++) {
				writer.write("\t"+genesID.get(i));
			}
			writer.write("\tID\n");
			for (int i = 0; i < header.length; i++) {
				//get Class name
				String className = "No";
				if (i>=borders[classIndex*2] && i<=borders[classIndex*2+1] ){
					className = "Yes";
				}
				if (className.equalsIgnoreCase("Yes")){
					writer.write(""+matrix[0][i]);
					for (int j = 1; j < matrix.length; j++) {
						writer.write("\t" + matrix[j][i]);
					}
					writer.write("\t"+className+"\n");
				}
				
			}
			for (int i = 0; i < header.length; i++) {
				//get Class name
				String className = "No";
				if (i>=borders[classIndex*2] && i<=borders[classIndex*2+1] ){
					className = "Yes";
				}
				if (className.equalsIgnoreCase("No")){
					writer.write(""+matrix[0][i]);
					for (int j = 1; j < matrix.length; j++) {
						writer.write("\t" + matrix[j][i]);
					}
					writer.write("\t"+className+"\n");
				}
				
			}
			writer.flush();
			writer.close();

		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		
		
		
	}
	
	public void printWekaCSVForRegression(String path){
		//print header
		
		BufferedWriter writer;
		try {
			writer = new BufferedWriter(new FileWriter(path));
			writer.write(genesID.get(0));
			for (int i = 1; i < genesID.size(); i++) {
				writer.write("\t"+genesID.get(i));
			}
			writer.write("\n");
			for (int i = 0; i < header.length; i++) {
				//get Class name
					writer.write(""+matrix[0][i]);
					for (int j = 1; j < matrix.length; j++) {
						writer.write("\t" + matrix[j][i]);
					}
					writer.write("\n");
			}
			
			writer.flush();
			writer.close();

		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		
	}
	
	public double[][] getDoubleTrasposedMatrix(){
		double[][] mat = new double[matrix[0].length][matrix.length];
		for (int i = 0; i < matrix.length; i++) {
			for (int j = 0; j < matrix[0].length; j++) {
				mat[j][i] = matrix[i][j];
			}
			
		}
		return mat;
	}
	
	public void printWekaCSVInstancesArePatients(String path,int[] borders,String[] classesNames){
		//print header
		
		BufferedWriter writer;
		try {
			writer = new BufferedWriter(new FileWriter(path));
			writer.write(genesID.get(0));
			for (int i = 1; i < genesID.size(); i++) {
				writer.write("\t"+genesID.get(i));
			}
			writer.write("\tID\n");
			for (int i = 0; i < header.length; i++) {
				writer.write(""+matrix[0][i]);
				for (int j = 1; j < matrix.length; j++) {
					writer.write("\t" + matrix[j][i]);
				}
				
				//get Class name
				String className = "";
				for (int classIndex = 0; classIndex < classesNames.length; classIndex++) {
					if (i>=borders[classIndex*2] && i<=borders[classIndex*2+1] ){
						className = classesNames[classIndex];
					}
				}
				writer.write("\t"+className+"\n");
			}
			writer.flush();
			writer.close();

		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		
		
		
	}
	
	public float[][] getMatrix() {
		return matrix;
	}
	public String[] getHeader() {
		return header;
	}
	public ArrayList<String> getGenesID() {
		return genesID;
	}
	
	public boolean containsGene(String name){
		return geneName2Index.containsKey(name);
	}
	
	public float geneGenesCorrelations(String first,String second){
		int i = geneName2Index.get(first);
		int j = geneName2Index.get(second);
		return staticMathHelperMethods.pearsonCorrelation(matrix[i], matrix[j]);
	}
	
	public void createGene2IndexHash(){
		for (int i = 0; i < genesID.size(); i++) {
			geneName2Index.put(genesID.get(i), i);
		}
	}
	
	public LinkedList<Integer> intensityFilterGenes(float thr,float percentage,int[] borders){
		
		LinkedList<Integer> goodGenes = new LinkedList<Integer>();
		
		for (int i = 0; i < matrix.length; i++) {
			boolean isGoodGene = true;
			for (int j = 0; j < borders.length; j+=2) {
				if (! isGoodGeneGeneByIntensities(thr, percentage,matrix[i],borders[j],borders[j+1])){
					isGoodGene = false;
					break;
				}
			}
			if (isGoodGene){
				goodGenes.add(i);
			}
			
		}
		
		return goodGenes;
	}
	
public LinkedList<Integer> intensityFilterGenesForOneClass(float thr,float percentage,int[] borders,int classIndex){
		
		LinkedList<Integer> goodGenes = new LinkedList<Integer>();
		
		for (int i = 0; i < matrix.length; i++) {
			
			if (isGoodGeneGeneByIntensities(thr, percentage,matrix[i],borders[classIndex],borders[classIndex+1])){
				goodGenes.add(i);
				
			}
		}
		
		return goodGenes;
	}
	
	
	private boolean isGoodGeneGeneByIntensities(float t,float p,float[] arr,int from,int to){
		
		float avg = staticMathHelperMethods.Average(arr, from, to);
		float sd = staticMathHelperMethods.sStatistic(arr, avg, from, to);
		if (sd == 0){
			System.out.println("removed a gene with 0 variation in class");
			return false;
		}
		
		float counter = 0;
		
		for (int i = from; i <= to ; i++) {
			if (arr[i]>=t){
				counter++;
			}
		}
		float n = to-from+1;
		
		if ((counter/n) >= p){
			return true;
		}
		
		return false;
	}
	
	public void addRowsToMatrix(ExpressionMatrix other){
		float[][] newMat  = new float[matrix.length+other.matrix.length][matrix[0].length];
		for (int i = 0; i < matrix.length; i++) {
			for (int j = 0; j < matrix[0].length; j++) {
				newMat[i][j] = matrix[i][j];
			}
		}
		
		int pointer = matrix.length;
		for (int i = 0; i < other.getMatrix().length; i++) {
			for (int j = 0; j < other.getMatrix()[0].length; j++) {
				newMat[pointer][j] = other.getMatrix()[i][j];
			}
			pointer++;
			genesID.add(other.getGenesID().get(i));
		}
		
		matrix = newMat;
	}
	
	public void setMatrix(float[][] matrix) {
		this.matrix = matrix;
	}
	
	public void floorData(float val){
		for (int i = 0; i < matrix.length; i++) {
			for (int j = 0; j < matrix[0].length; j++) {
				if (matrix[i][j] < val){
					matrix[i][j] = val;
				}
			}
		}
	
	}
	
	
	public void printMatrix(){
		System.out.print("ID");
		for (int i = 0; i < header.length; i++) {
			System.out.print("\t"+header[i]);
		}
		System.out.println();
		for (int i = 0; i < matrix.length; i++) {
			String currLine = genesID.get(i); 
			for (int j = 0; j < matrix[0].length; j++) {
				currLine += "\t" + matrix[i][j];
			}
			System.out.println(currLine);
		}
	}
	
	public void printMatrix(BufferedWriter writer) throws IOException{
		writer.write("ID");
		for (int i = 0; i < header.length; i++) {
			writer.write("\t"+header[i]);
		}
		writer.write("\n");
		for (int i = 0; i < matrix.length; i++) {
			String currLine = genesID.get(i); 
			for (int j = 0; j < matrix[0].length; j++) {
				currLine += "\t" + matrix[i][j];
			}
			writer.write(currLine+"\n");
		}
	}
	
	
	public ExpressionMatrix removeGenes(LinkedList<Integer> toRemove){
		ArrayList<String> ids = new ArrayList<String>();
		int numGenes = this.matrix.length - toRemove.size();
		float[][] newMat = new float[numGenes][this.matrix[0].length];
		int genesCounter = 0;
		for (int i = 0; i < matrix.length; i++) {
			if (toRemove.contains(i)){
				continue;
			}
			ids.add(genesID.get(i));
			for (int j = 0; j < newMat[0].length; j++) {
				newMat[genesCounter][j] = matrix[i][j];
			}
			genesCounter++;
		}
		ExpressionMatrix result = new ExpressionMatrix(newMat, header, ids);
		return result;
	}
	
	public ExpressionMatrix getExpandedByGraph(short[][] graph){
		ArrayList<String> ids = new ArrayList<String>();
		
		// first get all genes with a degree > 1
		// and count the number of vertices in the graph
		LinkedList<Integer> genes = new LinkedList<Integer>();
		int numEdges = 0;
		for (int i = 0; i < graph.length; i++) {
			genes.add(i);
			for (int j = 0; j < i; j++) {
				if (i==j){
					continue;
				}
				if (graph[i][j] > (short) 0){
					numEdges++;
				}
			}
			
		}
	
		float[][] newMat = new float[(numEdges+genes.size())][matrix[0].length];
		int genesFeatures = 0;
		//first add all genes
		for (Iterator<Integer> iterator = genes.iterator(); iterator.hasNext();) {
			Integer currGeneIndex = (Integer) iterator.next();
			for (int i = 0; i < newMat[0].length; i++) {
				newMat[genesFeatures][i]=matrix[currGeneIndex][i]*matrix[currGeneIndex][i];
			}
			ids.add(genesID.get(currGeneIndex)+"_Squared");
			genesFeatures++;
		}
		
		numEdges = 0;
		for (int i = 0; i < graph.length; i++) {
			for (int j = 0; j < i; j++) {
				if (graph[i][j]>(short) 0){
					numEdges++;
					for (int j2 = 0; j2 < newMat[0].length; j2++) {
						newMat[genesFeatures][j2]=2*matrix[i][j2]*matrix[j][j2];
					}
					genesFeatures++;
					ids.add(genesID.get(i)+"And"+genesID.get(j));
				}
			}
		}
		
		ExpressionMatrix result = new ExpressionMatrix(newMat, header, ids);
		return result;
	}
	
	public void standardizeByClasses(int[] borders){
		for (int i = 0; i < borders.length; i+=2) {
			int from = borders[i];
			int to = borders[i+1];
			
			float[] avgs = new float[matrix.length];
			for (int j = 0; j < matrix.length; j++) {
				avgs[j]=staticMathHelperMethods.Average(matrix[j],from,to);
			}
			float[] SDs = new float[matrix.length];
			for (int j = 0; j < SDs.length; j++) {
				SDs[j]=staticMathHelperMethods.sStatistic(matrix[j], avgs[j], from, to);
			}
			
			for (int j = 0; j < matrix.length; j++) {
				if ((new Float(SDs[j]).isNaN()) || SDs[j]==0){
					for (int j2 = from; j2 <= to; j2++) {
						matrix[j][j2]=(matrix[j][j2]-avgs[j]);
					}
				}
				else{
					for (int j2 = from; j2 <= to; j2++) {
						matrix[j][j2]=(matrix[j][j2]-avgs[j])/SDs[j];
					}
				}
				
			}
		}
	}
	
	public void shuffleExpressions(){
		Random rand = new Random();
		for (int i = 0; i < matrix.length; i++) {
			int last = matrix[0].length-1;
			while (last > 1){
				int next = rand.nextInt(last+1);
				float tmp = matrix[i][last];
				matrix[i][last]=matrix[i][next];
				matrix[i][last]=tmp;
				last--;
			}
		}
	}
	
	public ExpressionMatrix deepClone(){
		float[][] mat = new float[matrix.length][matrix[0].length];
		for (int i = 0; i < mat.length; i++) {
			for (int j = 0; j < mat[0].length; j++) {
				mat[i][j]=matrix[i][j];
			}
		}
		
		return new ExpressionMatrix(mat, header, genesID);
	}
	
	

	// assume that the last variable is the classes assignments
//	public ExpressionMatrix wekaInstancesToExpressionMatrix(Instances inst){
//		
//		float[][] mat = new float[inst.numAttributes()-1][inst.numInstances()];
//		for (int i = 0; i < mat[0].length; i++) {
//			Instance nextInstance = inst.instance(i);
//			for (int j = 0; j < mat.length; j++) {
//				mat [j][i] = (new Double(nextInstance.value(j))).floatValue();
//			}
//		}
//		
//		return null;
//	}
	
	
	public ExpressionMatrix getShuffeledMatrixByColumns(Random rand){
		
		float[][] mat = new float[matrix.length][matrix[0].length];
		for (int i = 0; i < mat.length; i++) {
			for (int j = 0; j < mat[0].length; j++) {
				mat[i][j]=matrix[i][j];
			}
		}
		
		int pointer = matrix[0].length-1;
//		System.out.println("pointer is starting at " + pointer);
		while (pointer > 1){
			int next = rand.nextInt(pointer);
			for (int i = 0; i < matrix.length; i++) {
				float tmp=mat[i][next];
				mat[i][next]=mat[i][pointer];
				mat[i][pointer]=tmp;
			}
			pointer--;
		}
		
//		for (int i = 0; i < mat.length; i++) {
//			for (int j = 0; j < mat[0].length; j++) {
//				System.out.println(mat[i][j]);
//			}
//		}
		
		return new ExpressionMatrix(mat, header, genesID);
		
	}
	
}
