package scripts;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;

import main_scripts.TscoresCalculator;

/**
 * This scripts evaluates the significance of the differential correlation
 * within and between modules in a given solution.
 * 
 * * The input contains: (1) expression matrix, (2) classes file, (3) the
 * clustering solution file , and (4) a bit specifying if this is a solution of
 * DICER or another algorithm. * 1. A gene expression matrix file. This is a tab
 * delimited file in which the first row contains the condition names, and the
 * first column contains the gene names. For example, see the LungCancer.txt
 * file LungCancer.zip. 2. A "classes file". This file contains partitioning of
 * the conditions in the matrix to groups. We assume that conditions from the
 * same groups are adjacent in the matrix. This file is tab delimited, in each
 * row there are three columns: the start index of the current group
 * (enumeration of the conditions starts from zero),the end index, and the group
 * name. For example, see the classesFile.txt in LungCancer.zip 3. This is a
 * tab-delimited file. Each line is of the format gene_name \t cluster_name. 4.
 * 0 if the solution is not of DICER, 1 otherwise.
 * 
 * 
 * This script evaluates the fold-change in DC between clusters as compared to
 * random gene sets of the same size.
 * 
 * Note, the output of this script was further analyzed for the paper to give
 * preference only for "significant" module pairs of other algorithms.
 */

public class DifferentialClusteringSolutionSignificanceEstimator {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length < 4) {
			System.out
					.println("usage: <expression matrix><classes file><clustering solution><our sol?><optional: map>");
			System.exit(1);
		}

		// this object is used here for getting the correlations and not the
		// t-scores
		TscoresCalculator sims = new TscoresCalculator(args[0], args[1],
				"pearson");
		sims.claculateSimilarities();
		Hashtable<String, HashSet<String>> geneSets = new Hashtable<String, HashSet<String>>();

		BufferedReader reader;
		try {
			reader = new BufferedReader(new FileReader(args[2]));
			String currLine = "";
			while ((currLine = reader.readLine()) != null) {
				String[] parsedLine = currLine.split("\t");
				if (parsedLine.length < 2) {
					continue;
				}
				String setName = parsedLine[1];
				if (setName.equalsIgnoreCase("grey")) {
					continue;
				}
				String gene = parsedLine[0];

				if (!geneSets.containsKey(setName)) {
					HashSet<String> set = new HashSet<String>();
					set.add(gene);
					geneSets.put(setName, set);
				} else {
					geneSets.get(setName).add(gene);
				}

			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		// go over the gene sets and get the differential correlation scores
		calculateDiffStatistics(sims, geneSets);

		// within sample;
		Random rand = new Random();
		System.out.println("within difference sample");
		for (Iterator<String> iterator = geneSets.keySet().iterator(); iterator
				.hasNext();) {
			String set = (String) iterator.next();
			ArrayList<Integer> list = getSetAsArrayList(sims, geneSets.get(set));
			for (int i = 0; i < 200; i++) {
				int a = list.get(rand.nextInt(list.size()));
				int b = list.get(rand.nextInt(list.size()));
				float diff = sims.class2Correlations.get(0)[a][b]
						- sims.class2Correlations.get(1)[a][b];
				if (diff < 0) {
					diff *= -1;
				}
				System.out.print(diff + " ");

			}
		}
		System.out.println();
		
		if (args.length>4){
			try {
				reader = new BufferedReader(new FileReader(args[4]));
				String currLine = "";
				while ((currLine = reader.readLine()) != null) {
					String[] parsedLine = currLine.split("\t");
//					System.out.println(currLine);
					if (parsedLine.length < 2) {
						continue;
					}
					HashSet<String> a = geneSets.get(parsedLine[0]);
					HashSet<String> b = geneSets.get(parsedLine[2]);
					float f = getPairFchange(sims, a, b);
					System.out.println(parsedLine[0]+"\t"+ parsedLine[2]+"\t"+f);
				}
			} catch (FileNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
			System.exit(0);
		}

		System.out.println("**** pairs analysis ****");
		int ourSol = Integer.parseInt(args[3]);
		if (ourSol == 0) {
			System.out.println("between sample");
			calculateDiffStatisticsForPairs(sims, geneSets);
			System.out.println();
		} else {
			System.out.println("between sample");
			calculateDiffStatisticsForPairsForOurAlgo(sims, geneSets);
			System.out.println();
		}

	}

	public static void calculateDiffStatistics(TscoresCalculator sims,
			Hashtable<String, HashSet<String>> geneSets) {

		int c = 0;
		float sumOfFoldChanges = 0;
		float minFoldChange = Float.MAX_VALUE;
		float[] fchanges = new float[geneSets.keySet().size()];
		float[] cluster2avgDiff = new float[geneSets.keySet().size()];
		float[] cluster2Size = new float[geneSets.keySet().size()];
		for (Iterator<String> iterator = geneSets.keySet().iterator(); iterator
				.hasNext();) {
			String set = (String) iterator.next();

			cluster2Size[c] = geneSets.get(set).size();
			float n = cluster2Size[c];

			float score = getSumOfScores(sims, geneSets.get(set));

			cluster2avgDiff[c] = score / ((n - 1) * n / 2);
			// System.out.println("##### CALCULATING cluster significance #####");
			// System.out.println("Real score is: " + cluster2avgDiff[c]);

			// now create 200 random clusters of the same size and calculate
			// fold changes
			Random rand = new Random();
			float avg = 0;
			float[] currArr = new float[200];
			for (int i = 0; i < 200; i++) {
				HashSet<String> currSet = new HashSet<String>();
				int size = geneSets.get(set).size();
				for (int j = 0; j < size; j++) {
					String next = sims.data.genesID.get(rand
							.nextInt(sims.data.genesID.size() - 1));
					while (currSet.contains(next)) {
						next = sims.data.genesID.get(rand
								.nextInt(sims.data.genesID.size() - 1));
					}
					currSet.add(next);
				}

				score = getSumOfScores(sims, currSet);
				score /= ((n - 1) * n / 2);
				currArr[i] = score;
				// System.out.println("random cluster score: " + score);
				avg += score;
			}
			avg /= (float) 200;
			float fchange = cluster2avgDiff[c] / avg;

			// float aRand = staticHelperMethods.Average(currArr);
			// float saRand = staticHelperMethods.sStatistic(currArr, aRand);
			// float zscore = (cluster2avgDiff[c] - aRand) / saRand;
			// double pval = 1 - TailProbs.probNormalC((double) zscore);
			// System.out.println("realscore\t" + cluster2avgDiff[c] + "\tavg:"
			// + aRand + "\tsd:" + saRand + "\tpval: " + pval);
			//

			fchanges[c] = fchange;

			// System.out.println("*** Fold change: " + fchange);
			sumOfFoldChanges += fchange;
			if (fchange < minFoldChange) {
				minFoldChange = fchange;
			}

			c++;
		}

		Arrays.sort(fchanges);
		float fChaneThr = fchanges[0];
		if (fchanges.length >= 20) {
			fChaneThr = fchanges[fchanges.length - 10];
		}

		float avg = 0;
		float count = 0;
		for (int i = 0; i < fchanges.length; i++) {
			if (fchanges[i] >= fChaneThr) {
				avg += fchanges[i];
				count++;
				System.out.println(" accepted thr: " + fchanges[i]);
			}
		}
		avg /= count;
		System.out.println("average fchange: " + avg);

	}

	public static void calculateDiffStatistics(TscoresCalculator sims,
			LinkedList<LinkedList<String>> geneSets) {

		int c = 0;
		float sumOfFoldChanges = 0;
		float minFoldChange = Float.MAX_VALUE;
		float[] fchanges = new float[geneSets.size()];
		float[] cluster2avgDiff = new float[geneSets.size()];
		float[] cluster2Size = new float[geneSets.size()];
		for (Iterator<LinkedList<String>> iterator = geneSets.iterator(); iterator
				.hasNext();) {
			LinkedList<String> set = (LinkedList<String>) iterator.next();

			cluster2Size[c] = set.size();
			float n = cluster2Size[c];

			float score = getSumOfScores(sims, set);

			cluster2avgDiff[c] = score / ((n - 1) * n / 2);
			// System.out.println("##### CALCULATING cluster significance #####");
			// System.out.println("Real score is: " + cluster2avgDiff[c]);

			// now create 200 random clusters of the same size and calculate
			// fold changes
			Random rand = new Random();
			float avg = 0;
			float[] currArr = new float[200];
			for (int i = 0; i < 200; i++) {
				HashSet<String> currSet = new HashSet<String>();
				int size = set.size();
				for (int j = 0; j < size; j++) {
					String next = sims.data.genesID.get(rand
							.nextInt(sims.data.genesID.size() - 1));
					while (currSet.contains(next)) {
						next = sims.data.genesID.get(rand
								.nextInt(sims.data.genesID.size() - 1));
					}
					currSet.add(next);
				}

				score = getSumOfScores(sims, currSet);
				score /= ((n - 1) * n / 2);
				currArr[i] = score;
				// System.out.println("random cluster score: " + score);
				avg += score;
			}
			avg /= (float) 200;
			float fchange = cluster2avgDiff[c] / avg;

			fchanges[c] = fchange;

			// System.out.println("*** Fold change: " + fchange);
			sumOfFoldChanges += fchange;
			if (fchange < minFoldChange) {
				minFoldChange = fchange;
			}

			c++;
		}

		Arrays.sort(fchanges);
		float fChaneThr = fchanges[0];
		if (fchanges.length >= 20) {
			fChaneThr = fchanges[fchanges.length - 10];
		}

		float avg = 0;
		float count = 0;
		for (int i = 0; i < fchanges.length; i++) {
			if (fchanges[i] >= fChaneThr) {
				avg += fchanges[i];
				count++;
				System.out.println(" accepted thr: " + fchanges[i]);
			}
		}
		avg /= count;
		System.out.println("average fchange: " + avg);

	}

	/**
	 * This method go over all set pairs and calculate the fold change against
	 * average of random modules. The method returns a hash that contains the
	 * accepted pairs (max size = 20).
	 * 
	 * @param sims
	 * @param geneSets
	 */
	public static void calculateDiffStatisticsForPairs(TscoresCalculator sims,
			Hashtable<String, HashSet<String>> geneSets) {

		int c = 0;
		float sumOfFoldChanges = 0;
		float minFoldChange = Float.MAX_VALUE;

		Hashtable<String, Double> pair2fChange = new Hashtable<String, Double>();
		Hashtable<String, Integer> pair2size = new Hashtable<String, Integer>();
		Hashtable<String, HashSet<String>> pairs = new Hashtable<String, HashSet<String>>();

		ArrayList<String> clusterIds = new ArrayList<String>();
		clusterIds.addAll(geneSets.keySet());
		int num = clusterIds.size();
		float[] fchanges = new float[num * (num - 1) / 2];

		for (int i = 0; i < clusterIds.size(); i++) {
			HashSet<String> set1 = geneSets.get(clusterIds.get(i));
			for (int j = 0; j < i; j++) {
				HashSet<String> set2 = geneSets.get(clusterIds.get(j));

				String currName = clusterIds.get(i) + "_" + clusterIds.get(j);

				HashSet<String> currPair = new HashSet<String>();
				currPair.addAll(set1);
				currPair.addAll(set2);
				pairs.put(currName, currPair);

				float realscore = getSumOfScores(sims, set1, set2, true);

				pair2size.put(currName, set1.size() + set2.size());

				Random rand = new Random();
				float avg = 0;
				float[] currArr = new float[200];

				for (int k = 0; k < 200; k++) {
					HashSet<String> currSet1 = new HashSet<String>();
					int size = set1.size();
					for (int l = 0; l < size; l++) {
						String next = sims.data.genesID.get(rand
								.nextInt(sims.data.genesID.size() - 1));
						while (currSet1.contains(next)) {
							next = sims.data.genesID.get(rand
									.nextInt(sims.data.genesID.size() - 1));
						}
						currSet1.add(next);
					}

					HashSet<String> currSet2 = new HashSet<String>();
					size = set2.size();
					for (int l = 0; l < size; l++) {
						String next = sims.data.genesID.get(rand
								.nextInt(sims.data.genesID.size() - 1));
						while (currSet2.contains(next)
								&& currSet1.contains(next)) {
							next = sims.data.genesID.get(rand
									.nextInt(sims.data.genesID.size() - 1));
						}
						currSet2.add(next);
					}

					float score = getSumOfScores(sims, currSet1, currSet2,
							false);
					currArr[i] = score;
					// System.out.println("random cluster score: " + score);
					avg += score;
				}
				avg /= (float) 200;
				float fchange = realscore / avg;
				pair2fChange.put(currName, (double) fchange);
				fchanges[c] = fchange;

				// System.out.println("*** Fold change: " + fchange);
				sumOfFoldChanges += fchange;
				if (fchange < minFoldChange) {
					minFoldChange = fchange;
				}

				c++;
			}
		}

		Arrays.sort(fchanges);
		float fChaneThr = fchanges[0];
		// System.out.println(fChaneThr);
		if (fchanges.length >= 20) {
			fChaneThr = fchanges[fchanges.length - 20];
		}

		float aveg = 0;
		Hashtable<String, HashSet<String>> accepted = new Hashtable<String, HashSet<String>>();
		for (Iterator<String> iterator = pairs.keySet().iterator(); iterator
				.hasNext();) {
			String setName = (String) iterator.next();
			HashSet<String> set = pairs.get(setName);
			if (accepted.keySet().size() >= 20) {
				break;
			}

			if (pair2fChange.get(setName) >= fChaneThr) {
				aveg += pair2fChange.get(setName);
				System.out.println("accepted fchange:\t"
						+ pair2fChange.get(setName));
				accepted.put(setName, set);
			}
		}
		aveg /= (float) accepted.keySet().size();
		System.out.println("average fold change of pairs: " + aveg);

	}

	public static void calculateDiffStatisticsForPairsForOurAlgo(
			TscoresCalculator sims, Hashtable<String, HashSet<String>> geneSets) {

		int c = 0;
		float sumOfFoldChanges = 0;
		float minFoldChange = Float.MAX_VALUE;

		Hashtable<String, Double> pair2fChange = new Hashtable<String, Double>();
		Hashtable<String, Integer> pair2size = new Hashtable<String, Integer>();
		Hashtable<String, HashSet<String>> pairs = new Hashtable<String, HashSet<String>>();

		ArrayList<String> clusterIds = new ArrayList<String>();
		clusterIds.addAll(geneSets.keySet());
		Collections.sort(clusterIds);

		int numUp = 0;
		int numDown = 0;
		for (Iterator<String> iterator = clusterIds.iterator(); iterator
				.hasNext();) {
			String string = (String) iterator.next();
			if (string.contains("up")) {
				numUp++;
			} else {
				numDown++;
			}
		}

		System.out.println("num up: " + numUp + ", num down: " + numDown);

		int num = clusterIds.size();
		float[] fchanges = new float[num * (num - 1) / 2];

		for (int i = 0; i < clusterIds.size(); i += 2) {
			String name1 = "";
			String name2 = "";

			if (i < numUp) {
				name1 = "up_" + i;
				name2 = "up_" + (i + 1);
			} else {
				int j = i - numUp;
				name1 = "down_" + j;
				name2 = "down_" + (j + 1);
			}
			HashSet<String> set1 = geneSets.get(name1);
			HashSet<String> set2 = geneSets.get(name2);

			String currName = name1 + "_" + name2;

			HashSet<String> currPair = new HashSet<String>();
			currPair.addAll(set1);
			currPair.addAll(set2);
			pairs.put(currName, currPair);

			float realscore = getSumOfScores(sims, set1, set2, true);

			pair2size.put(currName, set1.size() + set2.size());

			Random rand = new Random();
			float avg = 0;
			float[] currArr = new float[200];

			for (int k = 0; k < 200; k++) {
				HashSet<String> currSet1 = new HashSet<String>();
				int size = set1.size();
				for (int l = 0; l < size; l++) {
					String next = sims.data.genesID.get(rand
							.nextInt(sims.data.genesID.size() - 1));
					while (currSet1.contains(next)) {
						next = sims.data.genesID.get(rand
								.nextInt(sims.data.genesID.size() - 1));
					}
					currSet1.add(next);
				}

				HashSet<String> currSet2 = new HashSet<String>();
				size = set2.size();
				for (int l = 0; l < size; l++) {
					String next = sims.data.genesID.get(rand
							.nextInt(sims.data.genesID.size() - 1));
					while (currSet2.contains(next) && currSet1.contains(next)) {
						next = sims.data.genesID.get(rand
								.nextInt(sims.data.genesID.size() - 1));
					}
					currSet2.add(next);
				}

				float score = getSumOfScores(sims, currSet1, currSet2, false);
				currArr[i] = score;
				// System.out.println("random cluster score: " + score);
				avg += score;
			}
			avg /= (float) 200;
			float fchange = realscore / avg;
			pair2fChange.put(currName, (double) fchange);

			fchanges[c] = fchange;

			// System.out.println("*** Fold change: " + fchange);
			sumOfFoldChanges += fchange;
			if (fchange < minFoldChange) {
				minFoldChange = fchange;
			}

			c++;
		}

		Arrays.sort(fchanges);
		float fChaneThr = fchanges[0];
		// System.out.println(fChaneThr);
		// if (fchanges.length >= 20) {
		// fChaneThr = fchanges[fchanges.length - 20];
		// }
		// if (fChaneThr < 1) {
		// fChaneThr = 1;
		// }

		float aveg = 0;
		Hashtable<String, HashSet<String>> accepted = new Hashtable<String, HashSet<String>>();
		for (Iterator<String> iterator = pairs.keySet().iterator(); iterator
				.hasNext();) {
			String setName = (String) iterator.next();
			HashSet<String> set = pairs.get(setName);
			if (accepted.keySet().size() >= 20) {
				break;
			}
			aveg += pair2fChange.get(setName);
			if (pair2fChange.get(setName) > fChaneThr) {
				System.out.println("accepted fchange:\t"
						+ pair2fChange.get(setName));
				accepted.put(setName, set);
			}
		}
		aveg /= (float) accepted.keySet().size();
		System.out.println("average fold change of pairs: " + aveg);

	}

	public static float getPairFchange(
			TscoresCalculator sims, HashSet<String> a, HashSet<String> b) {

		float realscore = getSumOfScores(sims, a, b, false);

		Random rand = new Random();
		float avg = 0;
		float maxScore = 0;
		for (int k = 0; k < 200; k++) {
			HashSet<String> currSet1 = getRandomGeneSet(sims.data.genesID,a.size(),rand);
			HashSet<String> currSet2 = getRandomGeneSet(sims.data.genesID,b.size(),rand,a);
			float score = getSumOfScores(sims, currSet1, currSet2, false);
			avg += score;
			if (k == 0) {
				maxScore = score;
			} else {
				if (maxScore < score) {
					maxScore = score;
				}
			}
		}
		avg /= (float) 200;
//		float fchange1 = realscore / avg;
		float minFoldChane = realscore / maxScore;
		return minFoldChane;
	}

	public static float getSumOfScores(TscoresCalculator sims,
			Collection<String> geneSet) {
		ArrayList<Integer> l = getSetAsArrayList(sims, geneSet);
		float score = 0;
		for (int i = 0; i < l.size(); i++) {
			int a = l.get(i);
			for (int j = 0; j < i; j++) {
				int b = l.get(j);
				float diff = sims.class2Correlations.get(0)[a][b]
						- sims.class2Correlations.get(1)[a][b];
				if (diff < 0) {
					diff *= -1;
				}
				score += diff;
			}
		}
		return score;
	}

	public static float getSumOfScores(TscoresCalculator sims,
			HashSet<String> geneSet, HashSet<String> geneSet2, boolean print) {
		ArrayList<Integer> l1 = getSetAsArrayList(sims, geneSet);
		ArrayList<Integer> l2 = getSetAsArrayList(sims, geneSet2);

		float score = 0;
		float c = 0;

		for (int i = 0; i < l1.size(); i++) {
			int a = l1.get(i);
			for (int j = 0; j < l2.size(); j++) {
				int b = l2.get(j);
				if (l2 == l1) {
					continue;
				}
				c++;
				float diff = sims.class2Correlations.get(0)[a][b]
						- sims.class2Correlations.get(1)[a][b];
				if (diff < 0) {
					diff *= -1;
				}
				score += diff;
			}

		}

		Random rand = new Random();
		int counter = 0;
		while (counter < 20) {
			int a = l1.get(rand.nextInt(l1.size()));
			int b = l2.get(rand.nextInt(l2.size()));
			float diff = sims.class2Correlations.get(0)[a][b]
					- sims.class2Correlations.get(1)[a][b];
			if (diff < 0) {
				diff *= -1;
			}
			if (print) {
				System.out.print(diff + " ");
			}
			counter++;
		}

		return score / c;
	}

	public static ArrayList<Integer> getSetAsArrayList(TscoresCalculator sims,
			Collection<String> geneSet) {
		ArrayList<Integer> l = new ArrayList<Integer>();
		for (Iterator<String> iterator2 = geneSet.iterator(); iterator2
				.hasNext();) {
			String geneName = (String) iterator2.next();
			l.add(sims.data.genesID.indexOf(geneName));
		}
		return l;
	}
	
	public static HashSet<String> getRandomGeneSet(ArrayList<String> bg,int size,Random r){
		ArrayList<String> copy = new ArrayList<String>(bg);
		Collections.shuffle(copy, r);
		HashSet<String> s = new HashSet<String>();
		for (int i = 0; i < size; i++) {
			s.add(copy.get(i));
		}
		return s;
	}
	
	public static HashSet<String> getRandomGeneSet(ArrayList<String> bg,int size,Random r,HashSet<String> exclude){
		ArrayList<String> copy = new ArrayList<String>(bg);
		copy.removeAll(exclude);
		Collections.shuffle(copy, r);
		HashSet<String> s = new HashSet<String>();
		for (int i = 0; i < size; i++) {
			s.add(copy.get(i));
		}
		return s;
	}

}
