package scripts;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import scripts.HyperGeometricEnrichmentAnalysis;

import main_scripts.TscoresCalculator;
import math_utils.staticMathHelperMethods;

/**
 * This script is for evaluating a clustering solution.
 * 
 * The input contains: (1) expression matrix, (2) classes file, and (3) the clustering solution file.
 * 	 * 		1. A gene expression matrix file. This is a tab delimited file in which the first row
	 * 			contains the condition names, and the first column contains the gene names.
	 * 			For example, see the LungCancer.txt file LungCancer.zip.
	 * 		2. A "classes file". This file contains partitioning of the conditions in the matrix to groups.
	 * 			We assume that conditions from the same groups are adjacent in the matrix. 
	 * 			This file is tab delimited, in each row there are three columns: the start index of the 
	 * 			current group (enumeration of the conditions starts from zero),the end index, and the group name.
	 * 			For example, see the classesFile.txt in LungCancer.zip
	 * 		3. This is a tab-delimited file. Each line is of the format gene_name \t cluster_name.
 * 
 * Given a clustering solution (gene tab cluster) it calculates:
 * 
 * 1. Differential correlation statistics of the clusters:
 * Average absolute DC
 * Average weighted (by cluster size) absolute DC
 * Average cluster size
 * The number of the clusters in the solution
 * 
 * 2. Enrichment analysis results of the clusters (FDR 0.05).
 * To run this script two files should be in the running directory:
 * humanGenes2KeggNames.txt
 * Entrez2ProteinComplexes.txt
 * These files contain mapping of genes to functional terms (pathways or protein complexes).
 * Their format is the same as that of a clustering solution (gene tab term).
 * If these files are not in the running directory, an exception will appear.
 * In this case the output will contain the DC analysis results and the exception.
 * 
 * Other remarks:
 * 
 * A gene set named "grey" will be excluded from this analysis.
 * This cluster name may appear in DiffCoEx solution. The genes in this "cluster"
 * are genes that are to be excluded from DiffCoEx solution.
 * 
 * Unlike DifferentialClusteringSolutionSignificanceEstimator this script
 * is for analyzing the DC within clusters, without a comparison to random gene sets.
 */

public class DifferentialCorrelationOfClusteringAnalyzer {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length != 3){
			System.out.println("usage: <expression matrix><classes file><clustering solution>");
			System.out.println("calculating enrichments and average differential correlation scores for differential clustering solutions");
			System.out.println("we always assume that the class of interest is the firs (relvant to multiclass cases)");
			System.out.println("also we omit the cluster with the name grey");
			System.exit(1);
		}
		
		// this object is used here for getting the correlations and not the t-scores
		TscoresCalculator sims = new TscoresCalculator(args[0], args[1], "pearson");
		sims.claculateSimilarities();
		Hashtable<String,HashSet<String>> geneSets = new Hashtable<String, HashSet<String>>();
		
		BufferedReader reader;
		try {
			reader = new BufferedReader(new FileReader(args[2]));
			String currLine = "";
			while ((currLine=reader.readLine()) != null){
				String[] parsedLine = currLine.split("\t");
				if (parsedLine.length < 2){
					continue;
				}
				String setName = parsedLine[1];
				
				// in diffcoex solution, "grey" means all genes that are not assigned to modules.
				if (setName.equalsIgnoreCase("grey")){
					continue;
				}
				String gene = parsedLine[0];
				
				if (! geneSets.containsKey(setName)){
					HashSet<String> set = new HashSet<String>();
					set.add(gene);
					geneSets.put(setName, set);
				}
				else{
					geneSets.get(setName).add(gene);
				}
				
				
			}
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		// go over the gene sets and get the differential correlation scores
		float[] cluster2avgDiff = new float[geneSets.size()];
		float[] cluster2Size = new float[cluster2avgDiff.length];
		int c = 0;
		for (Iterator<String> iterator = geneSets.keySet().iterator(); iterator.hasNext();) {
			String set = (String) iterator.next();
			ArrayList<Integer> l = new ArrayList<Integer>();
			for (Iterator<String> iterator2 = geneSets.get(set).iterator(); iterator2.hasNext();) {
				String geneName = (String) iterator2.next();
				l.add(sims.data.genesID.indexOf(geneName));
			}
			cluster2Size[c] = l.size();
			float n = cluster2Size[c];
			float score = 0;
			for (int i = 0; i < l.size(); i++) {
				int a = l.get(i);
				for (int j = 0; j < i; j++) {
					int b = l.get(j);
					float diff = sims.class2Correlations.get(0)[a][b]-sims.class2Correlations.get(1)[a][b];
					if (diff < 0){
						diff *= -1;
					}
					score += diff;
				}
			}
			cluster2avgDiff[c] = score / ((n-1)*n/2);
			c++;
		}
		
		float[] weightedScore = new float[cluster2avgDiff.length];
		float sumWeightedScore = 0;
		for (int i = 0; i < weightedScore.length; i++) {
			weightedScore[i] = cluster2avgDiff[i]*cluster2Size[i];
			sumWeightedScore += weightedScore[i]; 
		}
		
		float avgDiff = staticMathHelperMethods.Average(cluster2avgDiff);
		float avgWeightedScore = staticMathHelperMethods.Average(weightedScore);
		float avgSize = staticMathHelperMethods.Average(cluster2Size);
		System.out.println("average DC\tweighted avg DV\taverage cluster size\tnum of sets");
		System.out.println(avgDiff + "\t" + avgWeightedScore + "\t" + sumWeightedScore + "\t" + avgSize + "\t" + geneSets.keySet().size());
		
		System.out.println("****** calculating enrichments *************");
		// perform enrichment analyses
		String keggPath = "humanGenes2KeggNames.txt";
		String complexesPath = "Entrez2ProteinComplexes.txt";
		System.out.println("****** seed enrichment results ******");
		System.out.println("###KEGG###");
		HyperGeometricEnrichmentAnalysis anal = new HyperGeometricEnrichmentAnalysis(
				keggPath, sims.data.getGenesID(), geneSets);
		anal.printResults(0.05);
		System.out.println("###Complexes###");
		anal = new HyperGeometricEnrichmentAnalysis(complexesPath,
				sims.data.getGenesID(), geneSets);
		anal.printResults(0.05);
	}

}
