package scripts;

import graph_algorithms.MetaModulePairMerger;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import scripts.HyperGeometricEnrichmentAnalysis;

import main_scripts.DICER_main;
import math_utils.ExpressionMatrix;

/**
 * This script calculates the "meta-file" of a clustering solution.
 * It was used, for example, to create the map of DC between modules.
 * This analysis gives an overview map of the DC instead of analyzing
 * specific gene pairs.
 * 
 * To use this script, you first need to save the graphs calculated by DICER.
 * This is done by giving an additional parameter to DICER (see the instructions
 * in DIECER_main).
 * 
 * The input is: the expression matrix, the path
 * in which the DICER graphs are kept, a clustering solution (i.e., gene to
 * gene sets mapping). For DICER analysis we recommend using the modules file
 * as we are interested here in a meta-graph of these modules.
 * 
 * Also, you can add an optinal input: gene to pathway mapping and FDR threshold.
 * When this input is given, we also output the enrichment analysis
 * results.
 * 
 * In the meta-graph we add an edge between tow modules in the 
 * sum of LLR scores between these modules in > 0. Also, in the sif file
 * we keep the "edge type". That is, if the edge denotes down correlation
 * or up-correlation.
 *
 */

public class GetMetaGraphFromSol {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length <3){
			System.out.println("usage: <data matrix><matrices path><clustering sol><Optional: classes for enrichment,FDR>");
			System.exit(1);
		}
		
		ExpressionMatrix mat = new ExpressionMatrix(args[0]);
		String pathName = args[1];
		float[][] dGGraph = DICER_main.readMatrix(pathName + "_Gdown.txt");
		float[][] uGGraph = DICER_main.readMatrix(pathName + "_Gup.txt");
		Hashtable<String, HashSet<String>> seeds = HyperGeometricEnrichmentAnalysis.readGeneSets(args[2], mat.genesID);
		Hashtable<String, HashSet<Integer>> seedsByIndex = new Hashtable<String, HashSet<Integer>>();
		for (Iterator<String> iterator = seeds.keySet().iterator(); iterator.hasNext();) {
			String seedId = (String) iterator.next();
			HashSet<String> set = seeds.get(seedId);
			HashSet<Integer> setInds = new HashSet<Integer>();
			for (Iterator<String> iterator2 = set.iterator(); iterator2.hasNext();) {
				String inSet = (String) iterator2.next();
				setInds.add(mat.genesID.indexOf(inSet));
			}
			seedsByIndex.put(seedId, setInds);
		}
		
		ArrayList<String> seedIds = new ArrayList<String>();
		if (args.length == 3){
			seedIds.addAll(seeds.keySet());
		}
		else{
			float fdr = Float.parseFloat(args[4]);
			HyperGeometricEnrichmentAnalysis enrichmentAnalysis = new HyperGeometricEnrichmentAnalysis(args[3], mat.genesID, seeds);
			enrichmentAnalysis.calcResults();
			enrichmentAnalysis.determineFDRThrIndependenceAssumption(fdr);
			seedIds.addAll(enrichmentAnalysis.enrichedClusters);
		}
		
	
		
		// print sif file
		for (int i = 0; i < seedIds.size(); i++) {
			HashSet<Integer> a = seedsByIndex.get(seedIds.get(i));
			for (int j = 0; j < i; j++) {
				HashSet<Integer> b = seedsByIndex.get(seedIds.get(j));
				double score = MetaModulePairMerger.getBetweenScores(a, b, dGGraph);
				if (score > 0){
					System.out.println(seedIds.get(i) +"\tdown\t" + seedIds.get(j));
				}
				else{
					score = MetaModulePairMerger.getBetweenScores(a, b, uGGraph);
					if (score > 0){
						System.out.println(seedIds.get(i) +"\tup\t" + seedIds.get(j));
					}
				}
				
			}
		}
		

	}

}
